#!/usr/bin/env python3

"""
Pre-requisite: pip install inotify

Brief: Use this script to launch a build script when source changes.

Usage: python assist.py <path to watch> <command to run>

Inotify (i.e. inotifywait) and other file change notification systems will
send many events all at once about how the file is being accessed or changed.
You can't you these raw events to trigger a build because you'd start running
the build prematurely.

This script is designed to watch for only mutation events (i.e. it ignores
opens, accesses, and close with nowrite). When using an editor like VSCode,
it is constantly accessing file meta data to determine if it needs to reread
files. We should not build everytime VSCode merely reads the metadata of a
file path.

The script waits for the inotify events to settle before running the build
again. For example, you may run a formatter that modifies many files in the
source directory. You don't want the build command to be executed after only
the first file has been written, you need to wait until all files have been
touched. This settlement is detected with a simple timeout. It waits for a
given duration of no events after an event before running the given reaction
command.

The script will not queue commands to run while changes are incoming. If the
given command takes a while to run, changes might be made to the source
while the command is running. The last change time is recorded and this is
checked against the last time the command was executed. If the difference is
greater than the given settlement duration it'll rerun the command.

The script has a signal handler to catch Ctrl-C. Ctrl-C is the expected way
to exit the watcher.py while its waiting for updates. I don't know what the
behavior of the system is if you Ctrl-C/SIGINT while the given user command
is being executed.

"""

import inotify.adapters
from threading import Thread, Lock
from time import monotonic

SETTLE_DURATION_DEFAULT = 1


class Context(object):
    def __init__(self, settle_duration=SETTLE_DURATION_DEFAULT):
        self._settle_duration = settle_duration

        self._event_lock = Lock()
        self._prev_time = 0
        self._last_checked = 0
        self._cur_time = 0

    @property
    def prev_time(self):
        return self._prev_time

    def update(self):
        self._event_lock.acquire()
        self._prev_time = monotonic()
        self._event_lock.release()

    def is_event(self):
        is_event = False
        self._event_lock.acquire()
        self._cur_time = monotonic()
        if self._prev_time != self._last_checked and self._cur_time - self._prev_time >= self._settle_duration:
            self._last_checked = self._prev_time
            is_event = True
        self._event_lock.release()
        return is_event


def inotifywait(ctx, path_to_watch):
    global prev_time
    i = inotify.adapters.InotifyTree(path_to_watch)

    for event in i.event_gen(yield_nones=False):
        (_, type_names, path, filename) = event

        # IGNORE: access, close_nowrite, open
        # WATCH: modify, attrib, close_write, close, moved_to, moved_from, move,
        #        move_self, create, delete, delete_self, unmount
        ignored_events = ['IN_ACCESS', 'IN_CLOSE_NOWRITE', 'IN_OPEN']
        for event_type in type_names:
            if event_type in ignored_events:
                break
        else:
            ctx.update()


def start_watching(path_to_watch, settle_duration=SETTLE_DURATION_DEFAULT):
    ctx = Context(settle_duration=settle_duration)

    inotify_thread = Thread(target=inotifywait, args=(
        ctx, path_to_watch), daemon=True)
    inotify_thread.start()

    return ctx



#!/usr/bin/env python3

import sys
import subprocess
import argparse

from signal import signal, SIGINT
from time import sleep

from thirdparty.builder.assist import start_watching


def KeyboardInterrupt_handler(signum, frame):
    sys.exit(0)


def main_loop(path_to_watch, command_args):
    ctx = start_watching(path_to_watch)

    no_command = False
    cmd = " ".join(command_args).strip()
    if len(cmd) == 0:
        no_command = True

    while True:
        sleep(1)
        is_event = ctx.is_event()

        if is_event:
            if no_command:
                print("CHANGE_EVENT_OBSERVED")
            else:
                subprocess.run(cmd, stderr=sys.stderr,
                               stdin=sys.stdin, stdout=sys.stdout, shell=True)


def main():

    # This must be here for the setuptools console_script to pick it up.
    signal(SIGINT, KeyboardInterrupt_handler)

    parser = argparse.ArgumentParser(
        description='Run command when directory content changes.')
    parser.add_argument('-p', '--path', dest='path_to_watch',
                        default='.', metavar='PATH',
                        help='Path to watch')
    parser.add_argument('-c', dest='command', default='',
                        metavar='CMD', help='Command to run on path change.')
    # Note: This only exist for --help. We extract this before parse_args().
    parser.add_argument('--', dest='command_args',
                        metavar='ARG',
                        nargs='*', help='Command arguments to run on path change.')

    # Get the location of the sentinal.
    sentinal = 0
    try:
        sentinal = sys.argv.index('--')
    except:
        pass

    if sentinal >= 1:
        argv = sys.argv[1:sentinal]
    else:
        argv = sys.argv[1:]

    args = parser.parse_args(argv)

    command_args = [args.command]
    if sentinal >= 1:
        command_args.extend(sys.argv[sentinal + 1:])

    print(command_args)
    main_loop(args.path_to_watch, command_args)


if __name__ == '__main__':
    main()



from setuptools import setup

with open("README.md", "r") as fh:
    long_description = fh.read()

setup(
    name="thirdparty-builder-assist",
    version="0.1.1",
    description="Tool for automatically building source on save.",
    long_description=long_description,
    long_description_content_type="text/markdown",
    author="Vincent Agriesti",
    author_email="crazychenz@gmail.com",
    url="https://github.com/third-party-dev/builder.assist",
    # In reality this could be 3.3. ... keeping this 3.7 or greater means
    # we reserve the right to use monotonic_ns() in the future. This should
    # probably never be less than 3.6 for asyncio.
    python_requires='~=3.7',
    packages=[
        "thirdparty",
        "thirdparty.builder",
        "thirdparty.builder.assist"
    ],
    entry_points={
        'console_scripts': ['assist=thirdparty.builder.assist.cli:main'],
    },
    package_dir={"": "src"},
    install_requires=[
        "inotify"
    ],
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Environment :: Console",
        "Intended Audience :: Developers",
        "Operating System :: POSIX :: Linux",
        "Topic :: Software Development :: Build Tools",
        "Topic :: Utilities",
    ],
)
